from ._impl import (
    CANSparkMax,
    MotorType,
    CANError,
    LimitSwitch,
    LimitSwitchPolarity,
    ControlType,
)

try:
    from .version import __version__
except ImportError:  # pragma: nocover
    __version__ = "master"



import sphinxify

_annotations = {
    "short": "int",
    "int": "int",
    "unsigned int": "int",
    "uint8_t": "int",
    "uint16_t": "int",
    "int32_t": "int",
    "uint32_t": "int",
    "size_t": "int",
    "double": "float",
    "char": "str",
    "bool": "bool",
    "rev::CANError": "CANError",
    "std::vector<uint8_t>": "typing.List[int]",
    "std::string": "str",
    "void": "None",
}

# fmt: off

def _gen_check(pname, ptype, strict=False):
    SIGNED_CHECK = "isinstance({0}, int) and -1<<{1} <= {0} < 1<<{1}".format
    UNSIGNED_CHECK = "isinstance({0}, int) and 0 <= {0} < 1<<{1}".format

    SIGNED_SIZES = {
        # "byte": 8,
        # "int8_t": 8,
        "short": 16,
        "int16_t": 16,
        "int32_t": 32,
        "int64_t": 64,
    }
    UNSIGNED_SIZES = {
        "uint8_t": 8,
        "uint16_t": 16,
        "uint32_t": 32,
        "uint64_t": 64,
    }

    # TODO: This does checks on normal types, but if you pass a ctypes value
    #       in then this does not check those properly.

    if ptype == 'bool':
        return 'isinstance(%s, bool)' % pname

    elif ptype in ('float', 'double'):
        if strict:
            return 'isinstance(%s, float)' % pname
        else:
            return 'isinstance(%s, (int, float))' % pname

    #elif ptype is C.c_char:
    #    return 'isinstance(%s, bytes) and len(%s) == 1' % (pname, pname)
    #elif ptype is C.c_wchar:
    #    return 'isinstance(%s, str) and len(%s) == 1' % (pname, pname)
    #elif ptype is C.c_char_p:
    #    return "%s is None or isinstance(%s, bytes) or getattr(%s, '_type_') is _C.c_char" % (pname, pname, pname)
    #elif ptype is C.c_wchar_p:
    #    return '%s is None or isinstance(%s, bytes)' % (pname, pname)

    elif ptype in ('int', 'long'):
        return 'isinstance(%s, int)' % pname
    elif ptype in SIGNED_SIZES:
        return SIGNED_CHECK(pname, SIGNED_SIZES[ptype] - 1)

    elif ptype == 'size_t':
        return 'isinstance(%s, int)' % (pname)
    elif ptype in UNSIGNED_SIZES:
        return UNSIGNED_CHECK(pname, UNSIGNED_SIZES[ptype])

    elif ptype is None:
        return '%s is None' % pname
    
    elif ptype.startswith("rev::"):
        return 'isinstance(%s, %s)' % (pname, ptype.rsplit("::", 1)[1])

    else:
        # TODO: do validation here
        #return 'isinstance(%s, %s)' % (pname, type(ptype).__name__)
        return None

# fmt: on


def _to_annotation(ctypename):
    computed = ctypename.replace("&", "").strip()
    idx = computed.rfind("::")
    if idx != -1:
        computed = computed[idx + 2 :]
    return _annotations.get(ctypename, computed)


def header_hook(header, data):
    """Called for each header"""

    for e in header.enums:
        e["x_namespace"] = e["namespace"]


def public_method_hook(fn, data):
    # Ignore operators, move constructors, copy constructors
    if (
        fn["name"].startswith("operator")
        or fn.get("destructor")
        or (fn.get("constructor") and fn["parameters"][0]["name"] == "&")
    ):
        fn["data"] = {"ignore": True}
        return

    # Python exposed function name converted to camelcase
    x_name = fn["name"]
    x_name = x_name[0].lower() + x_name[1:]

    x_in_params = []
    x_out_params = []
    x_rets = []

    # Simulation assertions
    x_param_checks = []
    x_return_checks = []

    data = data.get(fn["name"])
    if data is None:
        # ensure every function is in our yaml
        print("WARNING:", fn["parent"]["name"], "method", fn["name"], "missing")
        data = {}
        # assert False, fn['name']

    param_defaults = data.get("defaults", {})
    param_override = data.get("param_override", {})

    for i, p in enumerate(fn["parameters"]):
        if p["name"] == "":
            p["name"] = "param%s" % i
        p["x_type"] = p.get("enum", p["raw_type"])
        p["x_callname"] = p["name"]

        # Python annotations for sim
        p["x_pyann_type"] = _to_annotation(p["x_type"])
        if "forward_declared" in p:
            p["x_pyann_type"] = repr(p["x_pyann_type"])
            fn["forward_declare"] = True
            fn["parent"]["has_fwd_declare"] = True
        p["x_pyann"] = "%(name)s: %(x_pyann_type)s" % p
        p["x_pyarg"] = 'py::arg("%(name)s")' % p

        if "default" in p:
            p["x_pyann"] += " = " + p["default"]
            p["x_pyarg"] += "=" + p["default"]

        if p["name"] in param_override:
            p.update(param_override[p["name"]])
            x_in_params.append(p)
        elif p["pointer"]:
            p["x_callname"] = "&%(x_callname)s" % p
            x_out_params.append(p)
        elif p["array"]:
            asz = p.get("array_size", 0)
            if asz:
                p["x_pyann_type"] = "typing.List[%s]" % _to_annotation(p["raw_type"])
                p["x_type"] = "std::array<%s, %s>" % (p["x_type"], asz)
                p["x_callname"] = "%(x_callname)s.data()" % p
            else:
                # it's a vector
                pass

            x_out_params.append(p)
        else:
            chk = _gen_check(p["name"], p["x_type"])
            if chk:
                x_param_checks.append("assert %s" % chk)
            x_in_params.append(p)

        p["x_type"] += "&" * p["reference"]
        p["x_decl"] = "%s %s" % (p["x_type"], p["name"])

    x_callstart = ""
    x_callend = ""
    x_wrap_return = ""

    # Return all out parameters
    x_rets.extend(x_out_params)

    # if the function has out parameters and if the return value
    # is an error code, suppress the error code. This matches the Java
    # APIs, and the user can retrieve the error code from getLastError if
    # they really care
    if not len(x_rets) and fn["rtnType"] != "void":
        x_callstart = "auto __ret ="
        x_rets.insert(
            0,
            dict(
                name="__ret",
                x_type=fn["rtnType"],
                x_pyann_type=_to_annotation(fn["rtnType"]),
            ),
        )

        # Save some time in the common case -- set the error code to 0
        # if there's a single retval and the type is ErrorCode
        if fn["rtnType"] == "CANError":
            x_param_checks.append("retval = CANError.kOK")

    if len(x_rets) == 1 and x_rets[0]["x_type"] != "void":
        x_wrap_return = "return %s;" % x_rets[0]["name"]
        x_wrap_return_type = x_rets[0]["x_type"]
        x_pyann_ret = x_rets[0]["x_pyann_type"]
        chk = _gen_check("retval", x_wrap_return_type, strict=True)
        if chk:
            x_return_checks.append("assert %s" % chk)
    elif len(x_rets) > 1:
        x_pyann_ret = "typing.Tuple[%s]" % (
            ", ".join([p["x_pyann_type"] for p in x_rets]),
        )

        x_wrap_return = "return std::make_tuple(%s);" % ",".join(
            [p["name"] for p in x_rets]
        )
        x_wrap_return_type = "std::tuple<%s>" % (
            ", ".join([p["x_type"] for p in x_rets])
        )

        x_return_checks.append(
            "assert isinstance(retval, tuple) and len(retval) == %s" % len(x_rets)
        )
        for i, _p in enumerate(x_rets):
            chk = _gen_check("retval[%d]" % i, _p["raw_type"], strict=True)
            if chk:
                x_return_checks.append("assert %s" % chk)
    else:
        x_pyann_ret = "None"
        x_wrap_return_type = "void"

    # Temporary values to store out parameters in
    x_temprefs = ""
    if x_out_params:
        x_temprefs = ";".join(["%(x_type)s %(name)s" % p for p in x_out_params]) + ";"

    if "return" in data.get("code", ""):
        raise ValueError("%s: Do not use return, assign to retval instead" % fn["name"])

    # Rename internal functions
    if data.get("internal", False):
        x_name = "_" + x_name
    elif data.get("rename", False):
        x_name = data["rename"]
    elif fn["constructor"]:
        x_name = "__init__"

    if "doc" in data:
        doc = data["doc"]
    elif "doxygen" in fn:
        # work around a CppHeaderParser bug
        doc = fn["doxygen"].rpartition("*//*")[2]
        doc = sphinxify.process_raw(doc)

    if "hook" in data:
        eval(data["hook"])(fn, data)

    name = fn["name"]
    # XXX this probably isn't correct...
    x_returns = fn["namespace"] + fn["returns"]

    hascode = "code" in data or "get" in data or "set" in data

    # lazy :)
    fn.update(locals())


def class_hook(cls, data):
    # work around CppHeaderParser hoisting structs nested in classes to top
    if cls["parent"] is not None:
        cls["data"] = {"ignore": True}
        return

    data = data.get("data", {})
    data = data.get(cls["name"])
    if data is None:
        print("WARNING: class", cls["name"], "missing")
        data = {}

    # fix enum paths
    for e in cls["enums"]["public"]:
        e["x_namespace"] = e["namespace"] + "::" + cls["name"] + "::"

    cls["data"] = data
    methods_data = data.get("methods", {})
    for fn in cls["methods"]["public"]:
        public_method_hook(fn, methods_data)


def pid_get(fn, data):
    data["code"] = (
        "assert 0 <= slotID <= 3\n"
        'retval = self._hal_data["pid%d_{}" % slotID]'.format(fn["name"][3:].lower())
    )


def pid_set(fn, data):
    data["code"] = (
        "assert 0 <= slotID <= 3\n"
        'self._hal_data["pid%d_{}" % slotID] = float({})'.format(
            fn["name"][3:].lower(), fn["parameters"][0]["name"]
        )
    )



#
# Much of this copied from https://github.com/pybind/python_example.git
#

import os
from os.path import dirname, exists, join
from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist
import shutil
import subprocess
import sys
import setuptools

rev_lib_version = "1.0.27"

setup_dir = dirname(__file__)
git_dir = join(setup_dir, ".git")
base_package = "rev"
version_file = join(setup_dir, base_package, "version.py")

# Automatically generate a version.py based on the git version
if exists(git_dir):
    p = subprocess.Popen(
        ["git", "describe", "--tags", "--long", "--dirty=-dirty"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = p.communicate()
    # Make sure the git version has at least one tag
    if err:
        print("Error: You need to create a tag for this repo to use the builder")
        sys.exit(1)

    # Convert git version to PEP440 compliant version
    # - Older versions of pip choke on local identifiers, so we can't include the git commit
    v, commits, local = out.decode("utf-8").rstrip().split("-", 2)
    if commits != "0" or "-dirty" in local:
        v = "%s.post0.dev%s" % (v, commits)

    # Create the version.py file
    with open(version_file, "w") as fp:
        fp.write("# Autogenerated by setup.py\n__version__ = '{0}'".format(v))
        fp.write("\n__rev_version__ = '%s'" % rev_lib_version)

if exists(version_file):
    with open(join(setup_dir, base_package, "version.py"), "r") as fp:
        exec(fp.read(), globals())
else:
    __version__ = "master"

with open(join(setup_dir, "README.rst"), "r") as readme_file:
    long_description = readme_file.read()


#
# pybind-specific compilation stuff
#


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, "-std=c++14"):
        return "-std=c++14"
    elif has_flag(compiler, "-std=c++11"):
        return "-std=c++11"
    else:
        raise RuntimeError(
            "Unsupported compiler -- at least C++11 support " "is needed!"
        )


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {"msvc": ["/EHsc"], "unix": []}

    if sys.platform == "darwin":
        c_opts["unix"] += ["-stdlib=libc++", "-mmacosx-version-min=10.7"]

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == "unix":
            opts.append('-DVERSION_INFO="%s"' % rev_lib_version)
            opts.append("-s")  # strip
            opts.append("-g0")  # remove debug symbols
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append('/DVERSION_INFO=\\"%s\\"' % rev_lib_version)
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


install_requires = ["wpilib>=2019.0.0,<2020.0.0"]


class Downloader:
    """
        Utility object to allow lazily retrieving needed artifacts on demand,
        instead of distributing several extra MB with the pypi build.
    """

    def __init__(self):
        self._hallib = None
        self._halsrc = None
        self._revsrc = None
        self._wpiutillib = None
        self._wpiutilsrc = None

        rev_devdir = os.environ.get("RPY_REV_DEVDIR")
        if rev_devdir:
            # development use only -- preextracted files so it doesn't have
            # to download it over and over again
            # -> if the directory doesn't exist, it will download the current
            #    files to that directory

            self._hallib = join(rev_devdir, "hallib")
            self._halsrc = join(rev_devdir, "halsrc")
            self._revsrc = join(rev_devdir, "rev")
            self._wpiutillib = join(rev_devdir, "wpiutillib")
            self._wpiutilsrc = join(rev_devdir, "wpiutilsrc")

    # copy/paste from hal_impl.distutils
    def _download(self, url):
        import atexit
        import posixpath
        from urllib.request import urlretrieve, urlcleanup
        import sys

        print("Downloading", posixpath.basename(url))

        def _reporthook(count, blocksize, totalsize):
            percent = int(count * blocksize * 100 / totalsize)
            sys.stdout.write("\r%02d%%" % percent)
            sys.stdout.flush()

        filename, _ = urlretrieve(url, reporthook=_reporthook)
        atexit.register(urlcleanup)
        return filename

    def _download_and_extract_zip(self, url, to=None):
        import atexit
        import tempfile

        if to is None:
            # generate temporary directory
            tod = tempfile.TemporaryDirectory()
            to = tod.name
            atexit.register(tod.cleanup)

        zip_fname = self._download(url)
        return self._extract_zip(zip_fname, to)

    def _extract_zip(self, zip_fname, to):
        import shutil
        import zipfile

        with zipfile.ZipFile(zip_fname) as z:
            if isinstance(to, str):
                z.extractall(to)
                return to
            else:
                for src, dst in to.items():
                    with z.open(src, "r") as zfp:
                        with open(dst, "wb") as fp:
                            shutil.copyfileobj(zfp, fp)

    @property
    def hallib(self):
        if not self._hallib or not exists(self._hallib):
            import hal_impl.distutils

            self._hallib = hal_impl.distutils.extract_hal_libs(to=self._hallib)
        return self._hallib

    @property
    def halsrc(self):
        if not self._halsrc or not exists(self._halsrc):
            import hal_impl.distutils

            self._halsrc = hal_impl.distutils.extract_hal_headers(to=self._halsrc)
        return self._halsrc

    @property
    def wpiutillib(self):
        if not self._wpiutillib or not exists(self._wpiutillib):
            import hal_impl.distutils

            self._wpiutillib = hal_impl.distutils.extract_wpiutil_libs(
                to=self._wpiutillib
            )
        return self._wpiutillib

    @property
    def wpiutilsrc(self):
        if not self._wpiutilsrc or not exists(self._wpiutilsrc):
            import hal_impl.distutils

            url = "%s/%s/%s" % (
                hal_impl.distutils.wpiutil_site,
                hal_impl.distutils.wpiutil_version,
                "wpiutil-cpp-%s-headers.zip" % hal_impl.distutils.wpiutil_version,
            )

            self._wpiutilsrc = self._download_and_extract_zip(url, to=self._wpiutilsrc)
        return self._wpiutilsrc

    @property
    def revsrc(self):
        if not self._revsrc or not exists(self._revsrc):
            # Download and extract three libs
            base = "https://www.revrobotics.com/content/sw/max/sdk/maven/com/revrobotics/frc/"
            dirs = [
                "SparkMax-cpp/%(version)s/SparkMax-cpp-%(version)s-headers.zip",
                "SparkMax-cpp/%(version)s/SparkMax-cpp-%(version)s-linuxathenastatic.zip",
            ]

            for l in dirs:
                url = base + (l % dict(version=rev_lib_version))
                self._revsrc = self._download_and_extract_zip(url, to=self._revsrc)

        return self._revsrc


get = Downloader()

_travis_build = os.environ.get("TRAVIS_BUILD")

# Detect roboRIO.. not foolproof, but good enough
if exists("/etc/natinst/share/scs_imagemetadata.ini") or _travis_build:

    # Don't try to link when testing on travis, as it will fail
    # -> We can still catch compile errors, which is good enough I suspect
    if _travis_build:
        libraries = None
    else:
        libraries = ["wpiHal", "SparkMax"]

    wpilibc = join(setup_dir, base_package, "_impl", "wpilibc")

    ext_modules = [
        Extension(
            "rev._impl.rev_roborio",
            ["rev/_impl/rev_roborio.cpp"]
            + [
                join(wpilibc, "frc", f)
                for f in [
                    "CAN.cpp",
                    "DriverStation.cpp",
                    "Error.cpp",
                    "ErrorBase.cpp",
                    "RobotController.cpp",
                    "Timer.cpp",
                    "Utility.cpp",
                ]
            ],
            include_dirs=[
                # Path to pybind11 headers
                get_pybind_include(),
                get_pybind_include(user=True),
                get.revsrc,
                get.halsrc,
                wpilibc,
                join(get.halsrc),
                join(get.wpiutilsrc),
            ],
            libraries=libraries,
            library_dirs=[
                join(get.hallib, "linux", "athena", "shared"),
                join(get.wpiutillib, "linux", "athena", "shared"),
                join(get.revsrc, "linux", "athena", "static"),
            ],
            language="c++",
        )
    ]

    # This doesn't actually work, as it needs to be installed before setup.py is ran
    # ... but we specify it
    # install_requires = ['pybind11>=1.7']
    install_requires.append("robotpy-hal-roborio>=2019.0.0,<2020.0.0")
    cmdclass = {"build_ext": BuildExt}
else:
    install_requires.append("robotpy-hal-sim>=2019.0.0,<2020.0.0")
    ext_modules = None
    cmdclass = {}

#
# Autogenerating the required REV files is something that
# is done at sdist time. This means if you are testing builds,
# you have to run 'setup.py sdist build'.
#
# The advantage of doing it this way is that the autogen files
# are distributed with the pypi package, so simulation users
# don't have to install anything special to build this
#


class SDist(sdist):
    def run(self):
        from header2whatever import batch_convert

        # Do this before deleting the autogen directory, as it may fail
        revsrc = get.revsrc

        config_path = join(setup_dir, "gen", "gen.yml")
        outdir = join(setup_dir, "rev", "_impl", "autogen")

        shutil.rmtree(outdir, ignore_errors=True)

        batch_convert(config_path, outdir, revsrc)

        with open(join(outdir, "__init__.py"), "w"):
            pass

        super().run()


cmdclass["sdist"] = SDist

if os.environ.get("READTHEDOCS", None) == "True":
    sys.argv.insert(1, "sdist")

setup(
    name="robotpy-rev",
    version=__version__,
    author="Dustin Spicuzza",
    author_email="dustin@virtualroadside.com",
    url="https://github.com/robotpy/robotpy-rev",
    description="RobotPy bindings for REV third party libraries",
    long_description=long_description,
    packages=find_packages(),
    ext_modules=ext_modules,
    install_requires=install_requires,
    cmdclass=cmdclass,
    zip_safe=False,
    entry_points={
        "robotpylib": ["info = rev._impl.info:Info"],
        # "robotpysim": ["rev = rev._impl.sim_ui:RevUI"],
    },
)
